package checkers;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

import search.core.AIReflector;

import java.util.Set;

@SuppressWarnings("serial")
public class Checkers extends JFrame {
    
    private JButton startGame, evaluate, evalSearch, computerMove;
    private JComboBox<String> evalFuncs, searchers;
    private CheckersPanel board;
    private JLabel message, value;
    private EvalFunc evaluation;
    private Searcher searchAlg;
    private JTextField depthField, nodeField, timeField;
    private AIReflector<EvalFunc> evalFuncClasses;
    private AIReflector<Searcher> searcherClasses;
    
    private static boolean debug = false;
    
    public Checkers() {
        setTitle("Checkers");
        setSize(500, 600);
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        
        Container pane = getContentPane();
        pane.setLayout(new BorderLayout());
        
        board = new CheckersPanel();
        board.addMouseListener(new CheckerListener());
        pane.add(board, BorderLayout.CENTER);
        
        JPanel top = new JPanel(new GridLayout(4,1));
        pane.add(top, BorderLayout.NORTH);
        
        JPanel buttons = new JPanel(new FlowLayout());
        top.add(buttons);
        
        startGame = new JButton("Start new game");
        startGame.addActionListener(new GameStarter());
        buttons.add(startGame);
        
        computerMove = new JButton("Computer move");
        computerMove.addActionListener(new ComputerMover());
        buttons.add(computerMove);
        
        message = new JLabel();
        buttons.add(message);
        
        JPanel heuristics = new JPanel(new FlowLayout());
        top.add(heuristics);
        
        heuristics.add(new JLabel("Evaluation function"));
    	evalFuncClasses = new AIReflector<>(EvalFunc.class, "checkers.evaluators");
        evalFuncs = new JComboBox<>();
        for (String evalFuncStr: evalFuncClasses.getTypeNames()) {
        	evalFuncs.addItem(evalFuncStr);
        }
        evalFuncs.setSelectedIndex(0);
        evalFuncs.addActionListener(new Loader());
        heuristics.add(evalFuncs);
        
        evaluate = new JButton("Evaluate board");
        evaluate.addActionListener(new Evaluator());
        heuristics.add(evaluate);
        
        value = new JLabel();
        heuristics.add(value);
        
        JPanel heuristics2 = new JPanel(new FlowLayout());
        top.add(heuristics2);
        
        heuristics2.add(new JLabel("Search Algorithm"));
    	searcherClasses = new AIReflector<>(Searcher.class, "checkers.searchers", EvalFunc.class);
        searchers = new JComboBox<>();
        for (String searcherStr: searcherClasses.getTypeNames()) {
        	searchers.addItem(searcherStr);
        }
        searchers.setSelectedIndex(0);
        searchers.addActionListener(new Loader());
        heuristics2.add(searchers);
        
        heuristics2.add(new JLabel("Depth"));
        depthField = new JTextField(3);
        depthField.addActionListener(new Loader());
        heuristics2.add(depthField);
        
        evalSearch = new JButton("Evaluate with search");
        evalSearch.addActionListener(new Evaluator());
        heuristics2.add(evalSearch);
        
        JPanel nodeInfo = new JPanel(new FlowLayout());
        nodeInfo.add(new JLabel("Nodes expanded"));
        nodeField = new JTextField(10);
        nodeInfo.add(nodeField);
        nodeInfo.add(new JLabel("Time"));
        timeField = new JTextField(10);
        nodeInfo.add(timeField);
        top.add(nodeInfo);
        
        restart();
    }
    
    private void makeMove(Move m) {
        if (m != null && !board.getBoard().gameOver()) {
            board.getBoard().move(m);
            board.repaint();
            changeTurn();
        }
    }
    
    private void loadAlgorithms() {
        searchAlg = Searcher.makeSearcher((String)searchers.getSelectedItem(), (String)evalFuncs.getSelectedItem(), evalFuncClasses, searcherClasses);
        if (searchAlg == null) {
            JOptionPane.showMessageDialog(null, "Could not load heuristics");
        } else {
            evaluation = searchAlg.getEvaluator();
            resetDepth();
        }
    }
    
    private void resetDepth() {
        try {
        	nodeField.setText("");
        	timeField.setText("");
            int limit = Integer.parseInt(depthField.getText());
            searchAlg.setDepthLimit(limit);
        } catch (NumberFormatException nfe) {
        } finally {
            depthField.setText(Integer.toString(searchAlg.getDepthLimit()));
        }    	
    }
    
    private void restart() {
        board.getBoard().newGame();
        board.repaint();
       	changeTurn();
        loadAlgorithms();
    }
    
    private class Loader implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            loadAlgorithms();
        }
    }
    
    private class GameStarter implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            restart();
        }
    }
    
    private class Evaluator implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            double v = 0;
            if (e.getSource() == evaluate) {
                v = evaluation.evaluate(board.getBoard());
            } else if (e.getSource() == evalSearch) {
                v = searchAlg.selectMove(board.getBoard()).getScore();
            }
            value.setText("Value: " + ((double)(int)(v * 1000)/1000));
        }
    }
    
    private class ComputerMover implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            if (!board.getBoard().gameOver()) {
            	resetDepth();
            	long start = System.currentTimeMillis();
            	Move choice = searchAlg.selectMove(board.getBoard()).getMove();
            	long duration = System.currentTimeMillis() - start;
            	double seconds = duration / 1000.0;
            	if (board.getBoard().getLegalMoves(board.getBoard().getCurrentPlayer()).contains(choice)) {
            		makeMove(choice);
            		nodeField.setText(Integer.toString(searchAlg.numNodesExpanded()));
            		timeField.setText(String.format("%4.2f s", seconds));
            	} else {
            		JOptionPane.showMessageDialog(null, "AI requested an illegal move: " + choice);
            	}
            }
        }
    }
    
    private class CheckerListener extends MouseAdapter {
        private int colStart, rowStart;
        private boolean awaitingRelease;
        
        public void mousePressed (MouseEvent e) {
            colStart = board.getCol(e.getX());
            rowStart = board.getRow(e.getY());
            awaitingRelease = !board.getBoard().noPieceAt(rowStart, colStart);
            if (!awaitingRelease) {
                JOptionPane.showMessageDialog(null, "Illegal move");
                if (debug) {
                    System.out.println("No piece is at row " + rowStart + 
                    " and column " + colStart);
                }
            }
        }
        
        public void mouseReleased (MouseEvent e) {
            if (awaitingRelease) {
                Set<Move> legal = board.getBoard().getCurrentPlayerMoves();
                Move mouseMove = new Move(board.getBoard(), rowStart, colStart,
                board.getRow (e.getY()), 
                board.getCol (e.getX()));
                
                if (legal.contains(mouseMove)) {
                    makeMove(mouseMove);
                } else {
                    JOptionPane.showMessageDialog(null, "Illegal move");
                }
            } 
        }
    }
    
    // Pre: board != null
    // Post: Causes message to correspond to board's belief about the turn
    private void changeTurn () {
        if (board.getBoard().playerWins(Checkerboard.RED)) {
            message.setText ("Red wins!");
        } else if (board.getBoard().playerWins(Checkerboard.BLACK)) {
            message.setText ("Black wins!");
        } else if (board.getBoard().isTurnFor(Checkerboard.BLACK)) {
            message.setText ("Black's turn");
        } else {
            message.setText ("Red's turn");
        }
    }
    
    public static void main(String[] args) {
        Checkers c = new Checkers();
        c.setVisible(true);
    }
}
